from importlib import reload
import utils
reload(utils)
import keras
from keras import layers
from keras.models import load_model
import model
reload(model)
from model import Model
from sklearn.preprocessing import MinMaxScaler, StandardScaler


class Autoencoder(Model):
    
    def __init__(self, config, logs_dir, models_dir, _id=1):
        self.logs_dir = logs_dir
        self.models_dir = models_dir
        self._id = _id
        self.dataset_name = config["dataset_" + str(self._id)]
        self.input_shape = config[self.dataset_name]["model"]["input_shape"]
        self.optimizer = config[self.dataset_name]["model"]["train"]["optimizer"]
        self.loss = config[self.dataset_name]["model"]["train"]["loss"]
        self.n_epochs = config[self.dataset_name]["model"]["train"]["num_epochs"]
        self.batchsize = config[self.dataset_name]["model"]["train"]["batch_size"]
        self.ae_type = config[self.dataset_name]["model"]["type"]
        self.num_encoder_layers = config[self.dataset_name]["model"]["num_encoder_layers"]
        self.num_decoder_layers = config[self.dataset_name]["model"]["num_decoder_layers"]
        self.encoder_layers_sizes = config[self.dataset_name]["model"]["encoder_layers_sizes"]
        self.autoencoder, self.encoder, self.decoder, self.encoding_size = self.build_model()
    
    def build_model(self):
        if self.ae_type == "CAE":
            input_img = keras.Input(shape=(self.input_shape[0], self.input_shape[1], self.input_shape[2]))

            x = layers.Conv2D(18, (3, 3), activation='relu', padding='same')(input_img)
            x = layers.MaxPooling2D((2, 2), padding='same')(x)
            x = layers.Conv2D(9, (3, 3), activation='relu', padding='same')(x)
            x = layers.MaxPooling2D((2, 2), padding='same')(x)
            x = layers.Conv2D(9, (3, 3), activation='relu', padding='same')(x)
            encoded = layers.MaxPooling2D((2, 2), padding='same')(x)

            # at this point the representation is (4, 4, 8) i.e. 128-dimensional

            x = layers.Conv2D(9, (3, 3), activation='relu', padding='same')(encoded)
            x = layers.UpSampling2D((2, 2))(x)
            x = layers.Conv2D(9, (3, 3), activation='relu', padding='same')(x)
            x = layers.UpSampling2D((2, 2))(x)
            x = layers.Conv2D(18, (3, 3), activation='relu')(x)
            x = layers.UpSampling2D((2, 2))(x)
            decoded = layers.Conv2D(1, (3, 3), activation='sigmoid', padding='same')(x)
            self.num_decoder_layers = 7
            
        elif self.ae_type == "DAE":
            input_img = keras.Input(shape=(self.input_shape[0],))

            #### DAE-144
#             encoded = layers.Dense(576, activation='relu')(input_img)
#             encoded = layers.Dense(288, activation='relu')(encoded)
#             encoded = layers.Dense(144, activation='relu')(encoded)

#             decoded = layers.Dense(288, activation='relu')(encoded)
#             decoded = layers.Dense(576, activation='relu')(decoded)
#             decoded = layers.Dense(self.input_shape[0], activation='sigmoid')(decoded)
            
            #### DAE-32
            encoded = layers.Dense(self.encoder_layers_sizes[0], activation='relu')(input_img)
            for i in range(1, self.num_encoder_layers):
                encoded = layers.Dense(self.encoder_layers_sizes[i], activation='relu')(encoded)

            decoded = layers.Dense(self.encoder_layers_sizes[-2], activation='relu')(encoded)
            for i in range(self.num_decoder_layers-3, -1, -1):
                decoded = layers.Dense(self.encoder_layers_sizes[i], activation='relu')(decoded)
            decoded = layers.Dense(self.input_shape[0], activation='sigmoid')(decoded)
            
        
        # Get Autoencoder
        autoencoder = keras.Model(input_img, decoded, name=self.ae_type + "_" + str(self._id))
        autoencoder.summary()
        with open(self.logs_dir + 'AE_' + str(self._id) + '.txt','w') as fh:
            autoencoder.summary(print_fn=lambda x: fh.write(x + '\n'))
        # Get Encoder
        encoder = keras.Model(input_img, encoded, name="Encoder_" + str(self._id))
        encoder.summary()
        # Get Decoder
        encoding_size = encoder.layers[-1].output_shape[1:]
        decoder_input = keras.Input(shape=(encoding_size))
        next_input = decoder_input
        for layer in autoencoder.layers[-self.num_decoder_layers:]:
            next_input = layer(next_input)
        decoder = keras.Model(inputs=decoder_input, outputs=next_input, name="Decoder_" + str(self._id))
        decoder.summary()

        autoencoder.compile(optimizer=self.optimizer, loss=self.loss, metrics=['accuracy'])
        return autoencoder, encoder, decoder, encoding_size
    
    def train(self, train, test, restore_models=False):
        if self.ae_type == "DAE":
            train = train.reshape(train.shape[0], -1)
            test = test.reshape(test.shape[0], -1)
        if not restore_models:
            self.history = self.autoencoder.fit(train, train,
                            epochs=self.n_epochs,
                            batch_size=self.batchsize,
                            shuffle=True,
                            validation_data=(test, test))
            
            self.autoencoder.save(self.models_dir + 'AE_' + str(self._id) +'.h5')
            self.encoder.save(self.models_dir + 'E_' + str(self._id) +'.h5')
            self.decoder.save(self.models_dir + 'D_' + str(self._id) +'.h5')
        else:
            self.autoencoder = load_model(self.models_dir + 'AE_' + str(self._id) +'.h5')
            self.encoder = load_model(self.models_dir + 'E_' + str(self._id) +'.h5')
            self.decoder = load_model(self.models_dir + 'D_' + str(self._id) +'.h5')
        return
    
    def extract_latent(self, train, scale=False):
#         if self.ae_type == "DAE":
#             train = train.reshape(train.shape[0], -1)
#             test = test.reshape(test.shape[0], -1)
        latent_train = self.encoder.predict(train)
#         latent_test = self.encoder.predict(test)
        print(latent_train.shape)
        latent_train = latent_train.reshape(latent_train.shape[0], -1)
#         latent_test = latent_test.reshape(latent_test.shape[0], -1)
        
        if scale:
            train_scaler = MinMaxScaler()
#             test_scaler = MinMaxScaler()
            train_scaler.fit(latent_train)
#             test_scaler.fit(latent_test)

            latent_train = train_scaler.transform(latent_train)
#             latent_test = test_scaler.transform(latent_test)

        print(latent_train.shape)
        num_models = latent_train.shape[1]
        print("Num Models:", num_models)
        return latent_train