from math import sqrt
import math, sys
import numpy as np


def calculate_psnr(img1, img2, max_val=1.0):
    # img1 and img2 have range [0, 1]
    img1 = img1.astype(np.float64)
    img2 = img2.astype(np.float64)
    mse = np.mean((img1 - img2)**2)
    if mse == 0:
        return float('inf')
    return 20 * math.log10(max_val / math.sqrt(mse))


def calc_metrics_all(true_test, constructed_ts, fname=sys.stdout):
    mse = (np.square(true_test - constructed_ts)).mean(axis=None)
    rmse = sqrt(mse)
    nrmse = rmse / np.mean(true_test)
    mae = (np.abs(true_test - constructed_ts)).mean(axis=None)
    m = len(true_test)
    true_test_for_mape = np.where(true_test==0, 1, true_test)
    mape = (np.sum(np.abs((constructed_ts-true_test_for_mape)/true_test_for_mape))/m)*100
    psnr = calculate_psnr(true_test, constructed_ts, max_val=1.0)
    with open(fname, 'w') as f:
        print("MSE: %f" % mse, file=f)
        print("RMSE: %f" % rmse, file=f)
        print("NRMSE: %f" % nrmse, file=f)
        print("MAE: %f" % mae, file=f)
        print("MAPE: %f" % mape, file=f)
        print("PSNR: %f" % psnr, file=f)

    print("MSE: %f" % mse)
    print("RMSE: %f" % rmse)
    print("NRMSE: %f" % nrmse)
    print("MAE: %f" % mae)
    print("MAPE: %f" % mape)
    print("PSNR: %f" % psnr)